/*
Copyright 2014 Marco Fantini
This file is part of DEAL.
DEAL is free software : you can redistribute it and / or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DEAL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DEAL.If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef SeqTree_H
#define SeqTree_H


#include <cstdio>
#include <cstring>
#include <string>
#include <iostream>
#include <ctime>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <stdint.h>
#include <sstream>
#include <map>
#include <vector>
#include <set>
//#include <omp.h>


struct Seqnode;
struct node;
struct Groupnode;
struct Groupnode2;
struct SeqnodeLRpointers;

struct SeqnodeLRpointers
{
	Seqnode *LEFT_r1;
	Seqnode *RIGHT_r1;

	Seqnode *LEFT_r2;
	Seqnode *RIGHT_r2;

	Seqnode *LEFT_R;
	Seqnode *RIGHT_R;
};
struct Seqnode
{
	unsigned long ID;
	Groupnode *Group_r1;
	Groupnode *Group_r2;
	Groupnode *Group_R;
	SeqnodeLRpointers *rp;
	Seqnode *next;
};
struct node
{
	char *Seq;
	uint16_t *Fpos; //Fork position
	node **Fork;//Sibling
//	omp_lock_t lock;
};
struct Groupnode
{
	char *DefSeq;
	unsigned long ID;
	uint16_t Depth;
	unsigned long count;
	unsigned long TotCount;
	Groupnode *SubGroup;
	Groupnode *next;
	Seqnode *protein;
};

struct DefaultParam{
	unsigned int r1start;
	unsigned int r1end;
	unsigned int r2start;
	unsigned int r2end;
	unsigned int sl;
	unsigned int r1length;
	unsigned int r2length;
};
struct Parameters{
	DefaultParam ReadBoundariers;
	DefaultParam Seed;
	DefaultParam ProtBound;
	int QThresh = 32;
	double erThresh = 1;
	unsigned long MaxSequences = 0;
	std::string iFileName = "";
	std::string erFileName = "";
	std::string er2FileName = "";
	bool vl = false;
	bool test = false;
	bool splitReads = true;
	bool protan = false;
	int r1frame = 3;
	int r2frame = 1;
	int R1frame = 0;
	int R2frame = 0;
};
struct ErrorRateClass{
	unsigned int nTile = 0;
	unsigned int nIndex = 0;
	unsigned int npos = 0;

	double ***ER = NULL;
	bool given = false;

	std::map<std::string, int> mapIndex;
	std::map<std::string, int> mapTile;
};
namespace SeqTree
{
	extern 	Seqnode *pSeqROOT_R;
	extern 	Seqnode *pSeqROOT_r1;
	extern 	Seqnode *pSeqROOT_r2;
	extern Seqnode *SeqROOT;
	extern Seqnode *oldSeqnode;

	extern node *ROOT_R;
	extern Groupnode *GROOT_R;
	extern node *ROOT_r1;
	extern Groupnode *GROOT_r1;
	extern node *ROOT_r2;
	extern Groupnode *GROOT_r2;	

	extern Groupnode *pGROOT_R;
	extern Groupnode *pGROOT_r1;
	extern Groupnode *pGROOT_r2;

	extern const unsigned char A;
	extern const unsigned char G;
	extern const unsigned char C;
	extern const unsigned char T;
	extern const unsigned char N;
	extern const unsigned char inc;

	extern char Decompile[256];
	extern char Mappa[256];
	extern char ***prot;

	extern char *seedSeq;
	extern char *wholeSeq;
	extern char *protSeq;


	extern std::map<std::pair< Groupnode *, Groupnode *>, Groupnode2> Group_R;

	extern std::map<unsigned long, unsigned long> ngroup_r1_2ndstep;
	extern std::map<unsigned long, unsigned long> ngroup_r2_2ndstep;
	extern std::map<unsigned long, unsigned long> ngroup_R_2ndstep;

	extern unsigned long TOTSeq;

	extern unsigned long TOTGroup_r1;
	extern unsigned long TOTGroup_r2;
	extern unsigned long TOTGroup_R;

	extern unsigned long TOTGroup_r1_2ndstep;
	extern unsigned long TOTGroup_r2_2ndstep;
	extern unsigned long TOTGroup_R_2ndstep;

	extern unsigned long TOTGroup_r1_prot;
	extern unsigned long TOTGroup_r2_prot;
	extern unsigned long TOTGroup_R_prot;



	extern std::map<unsigned long, unsigned long> ngroup_R;
	extern std::map<unsigned long, unsigned long> ngroup_r1;
	extern std::map<unsigned long, unsigned long> ngroup_r2;

	extern Parameters par;
	extern ErrorRateClass er;
	extern ErrorRateClass er2;

	extern unsigned long TOTSeq_R_nopSTOP;
	extern unsigned long TOTSeq_r1_nopSTOP;
	extern unsigned long TOTSeq_r2_nopSTOP;
	extern unsigned long TOTGroup_R_nopSTOP;
	extern unsigned long TOTGroup_r1_nopSTOP;
	extern unsigned long TOTGroup_r2_nopSTOP;

void CaricaMappa();
void LoadErrorRate(std::ifstream &inputFile);
void ParseSequence(std::string Quality, std::string Sequence, std::ofstream &outputFile, unsigned long *ID);
void parsefile();
uint16_t numfork(node *Origin);
uint16_t indexfork(node *Origin, uint16_t position);
uint16_t indexfork_newindex(node *Origin, uint16_t position);
void removenode(node *N);
void clonenode(node *destination, node *origin, uint16_t sl);
void addfork(node *N, node *TREE, uint16_t sl, uint16_t i, uint16_t dimension);
int insert(node *N, node *TREE, uint16_t sl, uint16_t i, uint16_t lastFp, int F);
void straighten(std::ofstream &oFile_group, std::ofstream &oFile_groupF, std::ofstream &oFile_groupR, std::ofstream &oFile_stat, std::ofstream &oFile_DF);
void weave(Seqnode *S1, Seqnode *S2, int F);
void parsefileforgroup();
void RecursiveGroupWrite(Groupnode *GN, uint16_t Depth, std::ofstream &oFile, int F);
//void RecursiveGroupWrite(Groupnode2 *GN2, uint16_t sl, uint16_t Depth, std::ofstream &oFile);
void RecursiveGroupSearch(char *cSeq, Seqnode *seq, Groupnode *GN, uint16_t Depth, int SL, int F);
//void RecursiveGroupSearch(Seqnode *seq, Groupnode2 *GN2, uint16_t Depth);
void MergeSequences(char *Sequence, char *Sequence2, int SL);
void CompileSequence(std::string Header, std::string Quality, std::string Sequence, Seqnode *seqN, unsigned int SL);
char CheckCompatibility(char *Sequence, char *Sequence2, int SL, int SL2);
void parseInput(int argc, char** argv);
bool has_only_digits(std::string s);
bool has_only_digits_float(std::string s);
void setgroup(Seqnode *seq, Groupnode *GN, int F);
void loadprot();
void ProteinParseSequence(Groupnode *GN, int F);
void proteinanalysis();
void translateseq(char* Sequence, char* destination, bool* pSTOPflag, int F);
};
#endif 
