/*
Copyright 2014 Marco Fantini
This file is part of DEAL.
DEAL is free software : you can redistribute it and / or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DEAL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DEAL.If not, see <http://www.gnu.org/licenses/>.
*/

#include "SeqTree.h"

using namespace std;

namespace SeqTree
{

void ParseSequence(string Quality, string Sequence, ofstream &outputFile, unsigned long *ID)
{

	unsigned int SL = Sequence.length();

	unsigned int k = 0;
	for (unsigned int i = par.Seed.r1start; i < par.Seed.r1end; ++i){
		if (i < SL) seedSeq[k] = Mappa[(unsigned int)Sequence[i]];
		else seedSeq[k] = 0;
			
		k++;
	};

	for (unsigned int i = par.Seed.r2start; i <par.Seed.r2end; ++i){
		if (i < SL) seedSeq[k] = Mappa[(unsigned int)Sequence[i]];
		else seedSeq[k] = 0;
			
		k++;
	};

	Seqnode *newSeqnode;
	SeqnodeLRpointers *new_rp;
	newSeqnode = (Seqnode *)malloc(sizeof(Seqnode));
	new_rp = (SeqnodeLRpointers *)malloc(sizeof(SeqnodeLRpointers));
	newSeqnode->rp = new_rp;

	newSeqnode->rp->RIGHT_R = newSeqnode;
	newSeqnode->rp->LEFT_R = newSeqnode;

	if (par.splitReads){
		newSeqnode->rp->RIGHT_r1 = newSeqnode;
		newSeqnode->rp->LEFT_r1 = newSeqnode;
		newSeqnode->rp->RIGHT_r2 = newSeqnode;
		newSeqnode->rp->LEFT_r2 = newSeqnode;
	};

	static Groupnode *oldGnode_R;
	static Groupnode *newGnode_R;
	static Groupnode *oldGnode_r1;
	static Groupnode *newGnode_r1;
	static Groupnode *oldGnode_r2;
	static Groupnode *newGnode_r2;

	newGnode_R = (Groupnode *)malloc(sizeof(Groupnode));
	newGnode_R->ID = TOTGroup_R;
	newGnode_R->Depth = 0;
	newGnode_R->count = 0; // mi serve dopo
	newGnode_R->DefSeq = NULL;// mi serve dopo
	newGnode_R->next = NULL;// mi serve dopo
	newGnode_R->SubGroup = NULL;// mi serve dopo
	newGnode_R->TotCount = 1;
	ngroup_R[1]++;
	++TOTGroup_R;

	if (par.splitReads){
		oldGnode_r1 = newGnode_r1;
		newGnode_r1 = (Groupnode *)malloc(sizeof(Groupnode));
		newGnode_r1->ID = TOTGroup_r1;
		newGnode_r1->Depth = 0;
		newGnode_r1->count = 0; // mi serve dopo
		newGnode_r1->DefSeq = NULL;// mi serve dopo
		newGnode_r1->next = NULL;// mi serve dopo
		newGnode_r1->SubGroup = NULL;// mi serve dopo
		newGnode_r1->TotCount = 1;
		ngroup_r1[1]++;
		++TOTGroup_r1;	
		oldGnode_r2 = newGnode_r2;
		newGnode_r2 = (Groupnode *)malloc(sizeof(Groupnode));
		newGnode_r2->ID = TOTGroup_r2;
		newGnode_r2->Depth = 0;
		newGnode_r2->count = 0; // mi serve dopo
		newGnode_r2->DefSeq = NULL;// mi serve dopo
		newGnode_r2->next = NULL;// mi serve dopo
		newGnode_r2->SubGroup = NULL;// mi serve dopo
		newGnode_r2->TotCount = 1;
		ngroup_r2[1]++;
		++TOTGroup_r2;
	}


	newSeqnode->next = NULL;
	newSeqnode->Group_R = newGnode_R;
	newSeqnode->Group_r1 = newGnode_r1; //se par.splitReads  false  equivalente ad una inizializzazione a NULL
	newSeqnode->Group_r2 = newGnode_r2;
	newSeqnode->ID = *ID;

	node *newnode_R = NULL;
	node *newnode_r1= NULL;
	node *newnode_r2 = NULL; 

	newnode_R = (node *)malloc(sizeof(node));
	newnode_R->Fpos = (uint16_t *)malloc(2 * sizeof(uint16_t)); //1+1
	newnode_R->Fork = (node **)malloc(2 * sizeof(node *));
	newnode_R->Seq = (char *)malloc((par.Seed.sl + 1) * sizeof(char));
	newnode_R->Fpos[0] = 9999;
	newnode_R->Fpos[1] = 0;
	newnode_R->Fork[0] = (node *)newSeqnode;
	newnode_R->Fork[1] = NULL;


	memcpy(newnode_R->Seq, seedSeq, par.Seed.sl * sizeof(char));
	newnode_R->Seq[par.Seed.sl] = 0;


	if (par.splitReads){
		newnode_r1 = (node *)malloc(sizeof(node));
		newnode_r1->Fpos = (uint16_t *)malloc(2 * sizeof(uint16_t)); //1+1
		newnode_r1->Fork = (node **)malloc(2 * sizeof(node *));
		newnode_r1->Seq = (char *)malloc((par.Seed.r1length + 1) * sizeof(char));
		newnode_r1->Fpos[0] = 9999;
		newnode_r1->Fpos[1] = 0;
		newnode_r1->Fork[0] = (node *)newSeqnode;
		newnode_r1->Fork[1] = NULL;

		memcpy(newnode_r1->Seq, seedSeq, par.Seed.r1length * sizeof(char));
		newnode_r1->Seq[par.Seed.r1length] = 0;


		newnode_r2 = (node *)malloc(sizeof(node));
		newnode_r2->Fpos = (uint16_t *)malloc(2 * sizeof(uint16_t)); //1+1
		newnode_r2->Fork = (node **)malloc(2 * sizeof(node *));
		newnode_r2->Seq = (char *)malloc((par.Seed.r2length + 1) * sizeof(char));
		newnode_r2->Fpos[0] = 9999;
		newnode_r2->Fpos[1] = 0;
		newnode_r2->Fork[0] = (node *)newSeqnode;
		newnode_r2->Fork[1] = NULL;

		memcpy(newnode_r2->Seq, seedSeq + par.Seed.r1length, par.Seed.r2length* sizeof(char));
		newnode_r2->Seq[par.Seed.r2length] = 0;
	};

	if (ROOT_R){
		insert(newnode_R, ROOT_R, par.Seed.sl, 0, 0, 0);
		if (par.splitReads)	{
			insert(newnode_r1, ROOT_r1, par.Seed.r1length, 0, 0, 1);
			insert(newnode_r2, ROOT_r2, par.Seed.r2length, 0, 0, 2);
		};

	};

	if (!ROOT_R) {
		SeqROOT = newSeqnode;
		ROOT_R = newnode_R;
		GROOT_R = newGnode_R;
		oldSeqnode = SeqROOT;
		oldGnode_R = newGnode_R;
		if (par.splitReads)	{
			ROOT_r1 = newnode_r1;
			ROOT_r2 = newnode_r2;
			GROOT_r1 = newGnode_r1;
			GROOT_r2 = newGnode_r2;
			oldGnode_r1 = newGnode_r1;
			oldGnode_r2 = newGnode_r2;

		};

	};
	if (newGnode_R == newSeqnode->Group_R){ // = if newSeqnode->Group_R->TotCount == 1
		oldGnode_R = newGnode_R;
		oldGnode_R->next = newGnode_R; 
	};
	if (par.splitReads)	{
		if (newGnode_r1 == newSeqnode->Group_r1){
			oldGnode_r1 = newGnode_r1;
			oldGnode_r1->next = newGnode_r1;
		};
		if (newGnode_r2 == newSeqnode->Group_r2){

			oldGnode_r2 = newGnode_r2;
			oldGnode_r2->next = newGnode_r2;
		};
	};

	if (newSeqnode != SeqROOT) oldSeqnode->next = newSeqnode; //la ROOT non deve nextare se stessa!  come le normali sequenze (che hanno next vuoto)
	oldSeqnode = newSeqnode;

};

void parsefile()
{
	ifstream File;
	string ifilename = par.iFileName;
	clock_t StartTime = clock();
	unsigned long Iterations = 0;

	File.open(ifilename, ios::in);
	cout << "input: " << ifilename << '\n';

	if (er.given){
		cout << "input: " << par.erFileName << '\n';
	}


	ofstream ofile;
	string ofilename;
	ofilename = string(ifilename);
	ofilename += "_Tree.txt";
	cout << "output: " << ofilename.c_str() << '\n';
	ofile.open(ofilename.c_str(), ios::out);

	ofstream ofile_group;
	string ofilename_group;
	ofilename_group = string(ifilename);
	ofilename_group += "_Group.txt";
	cout << "output: " << ofilename_group.c_str() << '\n';
	ofile_group.open(ofilename_group.c_str(), ios::out);

	ofstream ofile_r1_group;
	string ofilename_r1_group;
	ofstream ofile_r2_group;
	string ofilename_r2_group;

	if (par.splitReads){
		ofilename_r1_group = string(ifilename);
		ofilename_r1_group += "_Group_r1.txt";
		cout << "output: " << ofilename_r1_group.c_str() << '\n';
		ofile_r1_group.open(ofilename_r1_group.c_str(), ios::out);

		ofilename_r2_group = string(ifilename);
		ofilename_r2_group += "_Group_r2.txt";
		cout << "output: " << ofilename_r2_group.c_str() << '\n';
		ofile_r2_group.open(ofilename_r2_group.c_str(), ios::out);
	};

	ofstream ofile_stat;
	string ofilename_stat;
	ofilename_stat = string(ifilename);
	ofilename_stat += "_Stat.txt";
	cout << "output: " << ofilename_stat.c_str() << '\n';
	ofile_stat.open(ofilename_stat.c_str(), ios::out);

	ofstream ofile_DF;
	string ofilename_DF;
	ofilename_DF = string(ifilename);
	ofilename_DF += "_DF.tsv";
	cout << "output: " << ofilename_DF.c_str() << '\n' << '\n';
	ofile_DF.open(ofilename_DF.c_str(), ios::out);



//
//#pragma omp parallel num_threads(4)
//	{
//
//#pragma omp master
//		{
//			string Header;
//			string Header2;
//			string Sequence;
//			string Quality;
//			while (getline(File, Header)){
//				getline(File, Sequence);
//				getline(File, Header2);
//				getline(File, Quality);
//#pragma omp task
//				{
//					if (Sequence.length()){ ParseSequence(Quality, Sequence, /*ofile,*/ &Iterations); };
//#pragma omp critical
//					{
//						Iterations++;
//						if (!(Iterations % 10000)) cerr << Iterations << "\n";
//					}
//				}
//#pragma omp taskwait
//			};
////
//
//		}
//	}
//
//


	string Header;
	string Header2;
	string Sequence;
	string Quality;
	while (getline(File, Header) && ((!par.test) || (Iterations < par.MaxSequences)))
	{
		getline(File, Sequence);
		getline(File, Header2);
		getline(File, Quality);
		if (Sequence.length()){ ParseSequence(Quality, Sequence, ofile, &Iterations); };
		Iterations++;
		if (!(Iterations % 10000)) cerr << Iterations << "\r";
	};

	cerr << Iterations;
	if (Iterations - 1) cerr << " sequenze analizzate.\n"; else cerr << " sequenza analizzata.\n";
	clock_t EndTime = clock();
	cerr << (EndTime - StartTime) / 1000000. << " secondi impiegati. ";
	cerr << "(" << 1000000. * Iterations / (EndTime - StartTime) << " sequenze al secondo)\n";


	File.close();

	removenode(ROOT_R);//liberiamo la memoria
	if (par.splitReads)	removenode(ROOT_r1);
	if (par.splitReads)	removenode(ROOT_r2);

	cerr << "Calculating Groups and statistics...\r";

	std::ofstream ofNull;

	if (!par.splitReads)straighten(ofile_group, ofNull, ofNull, ofile_stat, ofile_DF);
	if (par.splitReads)	straighten(ofile_group, ofile_r1_group, ofile_r2_group, ofile_stat, ofile_DF);

	ofile.close();
	ofile_group.close();
	if (par.splitReads) ofile_r1_group.close();
	if (par.splitReads) ofile_r2_group.close();
	ofile_stat.close();
	ofile_DF.close();

	ngroup_R.clear();
	ngroup_r1.clear();
	ngroup_r2.clear();
	ROOT_R = NULL;
	ROOT_r1 = NULL;
	ROOT_r2 = NULL;
	TOTGroup_R = 0;
	TOTGroup_r1 = 0;
	TOTGroup_r2 = 0;
};
}