/*
Copyright 2014 Marco Fantini
This file is part of DEAL.
DEAL is free software : you can redistribute it and / or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DEAL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DEAL.If not, see <http://www.gnu.org/licenses/>.
*/


#include "SeqTree.h"


namespace SeqTree
{

	Seqnode *pSeqROOT_R = NULL;
	Seqnode *pSeqROOT_r1 = NULL;
	Seqnode *pSeqROOT_r2 = NULL;
	Seqnode *SeqROOT = NULL;
	Seqnode *oldSeqnode = NULL;

	node *ROOT_R = NULL;
	Groupnode *GROOT_R = NULL;
	node *ROOT_r1 = NULL;
	Groupnode *GROOT_r1 = NULL;
	node *ROOT_r2 = NULL;
	Groupnode *GROOT_r2 = NULL;
	Groupnode *pGROOT_R = NULL;
	Groupnode *pGROOT_r1 = NULL;
	Groupnode *pGROOT_r2 = NULL;

	const unsigned char A = 1;
	const unsigned char C = 2;
	const unsigned char G = 4;
	const unsigned char T = 8;
	const unsigned char N = 0x0f;
	const unsigned char inc = 0x10;
	char Mappa[256] = { 0 };
	char Decompile[256] = { 0 };
	char ***prot;
	unsigned long TOTSeq_R_nopSTOP = 0;
	unsigned long TOTSeq_r1_nopSTOP = 0;
	unsigned long TOTSeq_r2_nopSTOP = 0;
	unsigned long TOTGroup_R_nopSTOP = 0;
	unsigned long TOTGroup_r1_nopSTOP = 0;
	unsigned long TOTGroup_r2_nopSTOP = 0;
	Parameters par;
	ErrorRateClass er;
	ErrorRateClass er2;
	//DefaultParam Default_ReadBoundariers = { r1start, r1end, r2start, r2end, sl, R1lenght, R2lenght};
	DefaultParam Default_ReadBoundariers = { 0, 320, 320, 540, 540, 320, 220 };
	DefaultParam Default_Seed = { 280, 300, 470, 490,  40,  20,  20 };
	DefaultParam Default_CDRVLVH = { 265, 320, 450, 510, 115,  55,  60 };//R1= VL R2=VH
	DefaultParam Default_CDRVHVL = { 265, 320, 450, 510, 115,  55,  60 };//R1= VH R2=VL

	char *seedSeq = NULL;
	char *wholeSeq = NULL;
	char *protSeq = NULL;

	std::map<unsigned long, unsigned long> ngroup_r1_2ndstep;
	std::map<unsigned long, unsigned long> ngroup_r2_2ndstep;
	std::map<unsigned long, unsigned long> ngroup_R_2ndstep;
	//std::map<std::pair< Groupnode *, Groupnode *>, Groupnode2> Group_R;

	unsigned long TOTSeq = 0;
	unsigned long TOTGroup_r1 = 0;
	unsigned long TOTGroup_r2 = 0;
	unsigned long TOTGroup_R = 0;

	unsigned long TOTGroup_r1_2ndstep = 0;
	unsigned long TOTGroup_r2_2ndstep = 0;
	unsigned long TOTGroup_R_2ndstep = 0;

	unsigned long TOTGroup_r1_prot = 0;
	unsigned long TOTGroup_r2_prot = 0;
	unsigned long TOTGroup_R_prot = 0;

	std::map<unsigned long, unsigned long> ngroup_R;
	std::map<unsigned long, unsigned long> ngroup_r1;
	std::map<unsigned long, unsigned long> ngroup_r2;


	void CaricaMappa()
	{
		Mappa['A'] = A;
		Mappa['C'] = C;
		Mappa['G'] = G;
		Mappa['T'] = T;

		Decompile[A] = 'A';
		Decompile[C] = 'C';
		Decompile[G] = 'G';
		Decompile[T] = 'T';
		Decompile[N] = 'N';

		//for debugging it't not possible
		Decompile[A | T] = 'W';
		Decompile[C | G] = 'S';

		Decompile[A | C] = 'M';
		Decompile[G | T] = 'K';

		Decompile[A | G] = 'R';
		Decompile[C | T] = 'Y';

		Decompile[C | G | T] = 'B';
		Decompile[A | G | T] = 'D';
		Decompile[A | C | T] = 'H';
		Decompile[A | C | G] = 'V';
		//up to here

		Decompile[inc | A] = 'A';
		Decompile[inc | C] = 'C';
		Decompile[inc | G] = 'G';
		Decompile[inc | T] = 'T';

		Decompile[inc | A | T] = 'W';
		Decompile[inc | C | G] = 'S';

		Decompile[inc | A | C] = 'M';
		Decompile[inc | G | T] = 'K';

		Decompile[inc | A | G] = 'R';
		Decompile[inc | C | T] = 'Y';

		Decompile[inc | C | G | T] = 'B';
		Decompile[inc | A | G | T] = 'D';
		Decompile[inc | A | C | T] = 'H';
		Decompile[inc | A | C | G] = 'V';

		Decompile[0x1f] = 'N';

		/*
		A	adenine	A
		C	cytosine		C
		G	guanine			G
		T	thymine				T
		U	uracil				U
		W	weak	A			T
		S	strong		C	G
		M	amino	A	C
		K	keto			G	T
		R	purine	A		G
		Y	pyrimidine		C		T
		B	not A(B comes after A)		C	G	T
		D	not C(D comes after C)	A		G	T
		H	not G(H comes after G)	A	C		T
		V	not T(V comes after T and U)	A	C	G
		N or - any base(not a gap)	A	C	G	T
		*/
	};


	void LoadErrorRate(std::ifstream &File)
	{

		std::string Header;
		getline(File, Header);
		std::string line;
		std::vector<std::string> *parts;
		std::set<std::string> setIndex;
		std::set<std::string> setTile;
		std::set<int> setpos;

		std::map<std::tuple<unsigned int, unsigned int, unsigned int>, double> mapFile;

		char * pch = NULL;
		char * np = NULL;
		int iTile = 0;
		char *dup = _strdup(Header.c_str());
		pch = strtok_s(dup, "\t", &np);
		er.nTile = 0;
		while (pch != NULL)
		{
			if (iTile >= 3) if ((setTile.insert(std::string(pch))).second == true) er.mapTile.insert(std::pair<std::string, int>(std::string(pch), (er.nTile++)));
			pch = strtok_s(NULL, "\t", &np);
			iTile++;

		};
		free(dup);
		while (getline(File, line))// && Iterations < MaxSequences)
		{
			parts = new std::vector<std::string>;
			np = NULL;
			dup = _strdup(line.c_str());
			pch = strtok_s(dup, "\t", &np);
			while (pch != NULL)
			{
				(*parts).push_back(pch);
				pch = strtok_s(NULL, "\t", &np);

			};

			std::string sIndex = (*parts)[2];
			std::string spos = (*parts)[1];
			int ipos = atoi(spos.c_str()) -1;

			if ((setIndex.insert(sIndex)).second == true) er.mapIndex.insert({ sIndex, er.nIndex++ });
			if ((setpos.insert(ipos)).second == true) er.npos++;

			int iIndex = er.mapIndex[sIndex];

			for (unsigned int iTile = 0; iTile < er.nTile; iTile++) mapFile.insert({ std::make_tuple(iIndex, iTile, ipos), atof(((*parts)[iTile + 3]).c_str()) });
			free(dup);
		};

		er.ER = new double**[er.nIndex];
		for (unsigned int i = 0; i < er.nIndex; i++)
		{
			er.ER[i] = new double*[er.nTile];
			for (unsigned int j = 0; j < er.nTile; j++)
			{
				er.ER[i][j] = new double[er.npos];
				for (unsigned int k = 0; k < er.npos; k++)
				{
					er.ER[i][j][k] = mapFile[std::make_tuple(i, j, k)];
				}
			}
		}
	};

	void LoadErrorRate2(std::ifstream &File_r2)
	{

		std::string Header;
		getline(File_r2, Header);
		std::string line;
		std::vector<std::string> *parts;
		std::set<std::string> setIndex;
		std::set<std::string> setTile;
		std::set<int> setpos;

		std::map<std::tuple<unsigned int, unsigned int, unsigned int>, double> mapFile;

		char * pch = NULL;
		char * np = NULL;
		int iTile = 0;
		char *dup = _strdup(Header.c_str());
		pch = strtok_s(dup, "\t", &np);
		er2.nTile = 0;
		while (pch != NULL)
		{
			if (iTile >= 3) if ((setTile.insert(std::string(pch))).second == true) er2.mapTile.insert(std::pair<std::string, int>(std::string(pch), (er2.nTile++)));
			pch = strtok_s(NULL, "\t", &np);
			iTile++;

		};
		free(dup);
		while (getline(File_r2, line))// && Iterations < MaxSequences)
		{
			parts = new std::vector<std::string>;
			np = NULL;
			dup = _strdup(line.c_str());
			pch = strtok_s(dup, "\t", &np);
			while (pch != NULL)
			{
				(*parts).push_back(pch);
				pch = strtok_s(NULL, "\t", &np);

			};

			std::string sIndex = (*parts)[2];
			std::string spos = (*parts)[1];
			int ipos = atoi(spos.c_str()) -1;

			if ((setIndex.insert(sIndex)).second == true) er2.mapIndex.insert({ sIndex, er2.nIndex++ });
			if ((setpos.insert(ipos)).second == true) er2.npos++;

			int iIndex = er2.mapIndex[sIndex];

			for (unsigned int iTile = 0; iTile < er2.nTile; iTile++) mapFile.insert({ std::make_tuple(iIndex, iTile, ipos), atof(((*parts)[iTile + 3]).c_str()) });
			free(dup);
		};

		er2.ER = new double**[er2.nIndex];
		for (unsigned int i = 0; i < er2.nIndex; i++)
		{
			er2.ER[i] = new double*[er2.nTile];
			for (unsigned int j = 0; j < er2.nTile; j++)
			{
				er2.ER[i][j] = new double[er2.npos];
				for (unsigned int k = 0; k < er2.npos; k++)
				{
					er2.ER[i][j][k] = mapFile[std::make_tuple(i, j, k)];
				}
			}
		}
	};


	void parseInput(int argc, char** argv) {
		if (argc == 1 || !std::string(argv[1]).compare("-h") || !std::string(argv[1]).compare("--help")) {
			std::cout << "Usage: SeqTree [options]                                                        " << std::endl;
			std::cout << "       -i    : input FastQ file                                                 " << std::endl;	// Implemented!
			std::cout << "       -b    : define the read(s) boundaries             [default 0 320 320 540]" << std::endl;	// Implemented!
			std::cout << "               The read boundaries are needed to specify the region of interest," << std::endl;
			std::cout << "               excluding, for examples leftover index sequence or other         " << std::endl;
			std::cout << "               non-Vregion parts.                                               " << std::endl;
			std::cout << "               If only a single domain is present use -b X Y where X and Y are  " << std::endl;
			std::cout << "               the boundaries of the wanted region.                             " << std::endl;
			std::cout << "       -s    : define the seed intervals               [default 280 300 470 490]" << std::endl;	// Implemented!
			std::cout << "               The seed regions are the regions analyzed in the first half of   " << std::endl;
			std::cout << "               the program. Errors in these regions will not be corrected       " << std::endl;
			std::cout << "       -j    : ignore single read informations. analyze only the whole sequence." << std::endl;	// Implemented!		Need correction
			std::cout << "               Should be use in most setting. Ignore only if the complexity     " << std::endl;
			std::cout << "               information of the single V regions of the scFv is needed.       " << std::endl;
			std::cout << "               If ignored. long computing time are to be expected               " << std::endl;
			std::cout << "       -vl   : use if input sequence are not of fixed length. plese use -b to   " << std::endl;	// !		
			std::cout << "               define the maximum length of the sequences     [default -b 0 540]" << std::endl;
			std::cout << "               Should be use if the input sequences are not fixed-lenght strings" << std::endl;
			std::cout << "               It needs specified boundaries to work. Usefull for sigle domain  " << std::endl;
			std::cout << "               libraries or libraries where the whole sequence can be read.     " << std::endl;
			std::cout << "       -er   : input file table where to load lambda-phage derived error rate(%)" << std::endl;
			std::cout << "                                                                [default ER.txt]" << std::endl;	// Implemented!
			std::cout << "               the file should be in tab separated format like this:            " << std::endl;
			std::cout << "            oldCycle   Bp  Index   1101    1102    1103 ...[other tile names]..." << std::endl;
			std::cout << "            7          1   TCAGCG  0.08    0.06    0.07 ..                      " << std::endl;
			std::cout << "            8          2   TCAGCG  0.07    0.08    0.06 ..                      " << std::endl;
			std::cout << "            9          3   TCAGCG  0.09    0.1     0.09 ..                      " << std::endl;
			std::cout << "            10         4   TCAGCG  0.07    0.08    0.05 ..                      " << std::endl;
			std::cout << "            11         5   TCAGCG  0.09    0.08    0.11 ..                      " << std::endl;
			std::cout << "               where oldcycle is not important for the program but shold be used to keep track of the sequenced" << std::endl;
			std::cout << "       -test : analyze only the first X sequences             [default: 100 000]" << std::endl;	// Implemented!

			std::cout << std::endl;
			std::cout << "######################## THRESHOLDS #####################                       " << std::endl;
			std::cout << "       -Qt   : set a threshold to Phred quality score under which the base is   " << std::endl;
			std::cout << "               consider uncertain                                   [default 32]" << std::endl;	// Implemented!
			std::cout << "       -ert  : set a threshold to lambda-phage derived error rate under which   " << std::endl;
			std::cout << "               the base is consider uncertain                        [default 1]" << std::endl;	// Implemented!
			std::cout << std::endl;
			std::cout << "######################### PROTEINS ######################                       " << std::endl; 
			std::cout << "       -p    : analyze also the protein diversity (frame must be specified)     " << std::endl;	// Implemented!		Need correction
			std::cout << "       -pfr  : define the protein frames (stating from read boundaries, one for " << std::endl;	// Implemented!
			std::cout << "               each read)                                         [default 3 1] " << std::endl;
			std::cout << "               as an example if the frame for each reads is known to be 1 use   " << std::endl;
			std::cout << "               -pfr 1 1 . In the most common case this option is used in        " << std::endl;
			std::cout << "               conjunction with -vl in single domain libraries. Since there will" << std::endl;
			std::cout << "               be only one chain use -pfr X whre X is the desired frame.        " << std::endl;
			std::cout << std::endl;
			std::cout << "######################## DEPRECATED #####################                       " << std::endl;
			std::cout << "       -CDR  : seed is CDR                         [VHVL or VLVH. default: VHVL]" << std::endl;	// Implemented!		BUG when -vl enabled
			std::cout << "               DEPRECATED. could be use insetad of -s for seed in CDRs          " << std::endl;
			std::cout << "       -WHOLE: seed is the whole sequence                                       " << std::endl;	// Implemented!
			std::cout << "               DEPRECATED. could be use insetad of -s for setting the seed as   " << std::endl;
			std::cout << "               the whole sequence                                               " << std::endl;
			std::cout << std::endl;
			std::cout << "########################### HELP ########################                       " << std::endl;
			std::cout << "       -h    : show this help                                                   " << std::endl;	// Implemented!
			std::cout << "       --help: show this help                                                   " << std::endl;	// Implemented!
			std::cout << std::endl;
			std::cout << "######################### EXAMPLES ######################                       " << std::endl;
			std::cout << "Examples: SeqTree -i lib.fastq                                                  " << std::endl;
			std::cout << "          SeqTree -i lib.fastq -CDR                                             " << std::endl;
			std::cout << "          SeqTree -i lib.fastq -CDR VLVH                                        " << std::endl;
			std::cout << "          SeqTree -i lib.fastq -s 30 50 500 530                                 " << std::endl;
			std::cout << "          SeqTree -i lib.fastq -WHOLE                                           " << std::endl;
			std::cout << "          SeqTree -i lib.fastq -er er.txt                                       " << std::endl;
			exit(0);
			//linelen 80 "################################################################################"

			//std::cout << "       -t    : Number of threads [default 1]" << std::endl;																	// ToDo
			//std::cout << "       -o    : Output file [Default: output.dat]" << std::endl;																// ToDo
		}


		par.QThresh = 32;
		par.erThresh = 1;
		par.MaxSequences = 100000;
		par.iFileName = "";
		par.test = false;
		par.ReadBoundariers = Default_ReadBoundariers;
		par.Seed = Default_Seed;
		par.vl = false;
		bool fError = false;
		int Warning = 0;
		int seedset = 0;
		int finput = 0;
		bool fSetBound = false;

		for (int i = 0; i < argc; i++) {
			if (!std::string(argv[i]).compare("-i")) {
				finput++;
				std::cout << "Checking input file..." << std::flush;
				if (i + 1 <= argc) {
					par.iFileName = std::string(argv[i + 1]);
					std::ifstream File;
					File.open(par.iFileName, std::ios::in);
					if (!File)
					{
						std::cout << "ERROR: Failed to open \"" << par.iFileName << "\"." << std::endl;
						fError = true;
					}
					else {
						std::string str;
						if (!std::getline(File, str)) { // if problem in line 1 (file empty)
							std::cout << "ERROR: Input file conitain no character" << std::endl;
							fError = true;
						}
						else {
							if (std::getline(File, str)) { // if !problem in line 2 (error or eof)
								if (!fSetBound)
								{
									par.ReadBoundariers.r1start = 0;
									par.ReadBoundariers.r2end = (unsigned int)str.length();
									par.ReadBoundariers.sl = (unsigned int)str.length();
								}
							}
						};
						std::cout << "done." << std::endl;
					};
					File.close();
					i++;
				}
				else {
					std::cout << "ERROR: Missing file argument" << std::endl;
					fError = true;
				};
			}

			//if (!std::string(argv[i]).compare("-o")){
			//	par.oFileName = std::string(argv[i + 1]);
			//	i++;
			//}
			//if (!std::string(argv[i]).compare("-t")){
			//	params.nThreads = atoi(argv[i + 1]);
			//	i++;
			//}

			if (!std::string(argv[i]).compare("-CDR")) {
				std::cout << "Seed set to CDR." << std::flush;
				seedset++;
				if (i + 1 <= argc) {
					if (argv[i + 1][0] != ('-')) {
						i++;
						if (!std::string(argv[i]).compare("VHVL")) {
							std::cout << std::endl;
							par.Seed = Default_CDRVHVL;
						}
						else if (!std::string(argv[i]).compare("VLVH")) {
							std::cout << std::endl;
							par.Seed = Default_CDRVLVH;
						}
						else {
							par.Seed = Default_CDRVHVL;
							std::cout << " WARNING: \"" << std::string(argv[i]) << "\" is not a valid argument. Set to CDR default." << std::endl;
							Warning++;
						}
					}
					else {			// no argument for cdr: next is -xx
						std::cout << std::endl;
						par.Seed = Default_CDRVHVL;
					}
				}
				else {					// no argument for cdr: -CDR is the last argument
					std::cout << std::endl;
					par.Seed = Default_CDRVHVL;
				};
			}

			if (!std::string(argv[i]).compare("-WHOLE")) {
				std::cout << "Seed set to whole sequence." << std::endl;
				par.Seed = Default_ReadBoundariers;
				seedset++;
			}

			if (!std::string(argv[i]).compare("-s")) {
				seedset++;
				int done = 0;
				if (i + 4 <= argc) {
					if (has_only_digits(std::string(argv[i + 1])) && has_only_digits(std::string(argv[i + 2])) &&
						has_only_digits(std::string(argv[i + 3])) && has_only_digits(std::string(argv[i + 4]))) {
						DefaultParam tmps;
						tmps.r1start = atoi(argv[i + 1]);
						tmps.r1end = atoi(argv[i + 2]);
						tmps.r2start = atoi(argv[i + 3]);
						tmps.r2end = atoi(argv[i + 4]);
						tmps.r1length = tmps.r1end - tmps.r1start;
						tmps.r2length = tmps.r2end - tmps.r2start;
						tmps.sl = tmps.r1length + tmps.r2length;
						i = i + 4;
						par.Seed = tmps;
						std::cout << "Seed set to [" << par.Seed.r1start << ":" << par.Seed.r1end << "][" << par.Seed.r2start << ":" << par.Seed.r2end << "]" << std::endl;
						done++;
					}
				}
				if (!done & i + 2 <= argc) {
					if (has_only_digits(std::string(argv[i + 1])) && has_only_digits(std::string(argv[i + 2]))) {
						DefaultParam tmps;
						tmps.r1start = atoi(argv[i + 1]);
						tmps.r1end = atoi(argv[i + 2]);
						tmps.r2start = tmps.r1end;
						tmps.r2end = tmps.r1end;
						tmps.r1length = tmps.r1end - tmps.r1start;
						tmps.r2length = tmps.r2end - tmps.r2start;
						tmps.sl = tmps.r1length + tmps.r2length;
						i = i + 2;
						par.Seed = tmps;
						std::cout << "Seed set to [" << par.Seed.r1start << ":" << par.Seed.r1end << "]" << std::endl;

						par.splitReads = false;
						std::cout << "Only two parameter given. Sequence will be treated as a single read." << std::endl;
						done++;
					}
				}
				if (!done){
					par.Seed = Default_Seed;
					std::cout << " WARNING: missing arguments. Seed set to default." << std::endl;
					Warning++;
				};
			}

			if (!std::string(argv[i]).compare("-b")) {
				int done = 0;
				if (i + 4 <= argc) {
					if (has_only_digits(std::string(argv[i + 1])) && has_only_digits(std::string(argv[i + 2])) &&
						has_only_digits(std::string(argv[i + 3])) && has_only_digits(std::string(argv[i + 4]))) {
						DefaultParam tmps;
						tmps.r1start = atoi(argv[i + 1]);
						tmps.r1end = atoi(argv[i + 2]);
						tmps.r2start = atoi(argv[i + 3]);
						tmps.r2end = atoi(argv[i + 4]);
						tmps.r1length = tmps.r1end - tmps.r1start;
						tmps.r2length = tmps.r2end - tmps.r2start;
						tmps.sl = tmps.r1length + tmps.r2length;
						i = i + 4;
						par.ReadBoundariers = tmps;
						std::cout << "Boundaries set to [" << par.ReadBoundariers.r1start << ":" << par.ReadBoundariers.r1end << "][" << par.ReadBoundariers.r2start << ":" << par.ReadBoundariers.r2end << "]" << std::endl;
						fSetBound = true;
						done++;
					}
				}
				if (!done & i + 2 <= argc) {
					if (has_only_digits(std::string(argv[i + 1])) && has_only_digits(std::string(argv[i + 2]))) {
						DefaultParam tmps;
						tmps.r1start = atoi(argv[i + 1]);
						tmps.r1end = atoi(argv[i + 2]);
						tmps.r2start = tmps.r1end;
						tmps.r2end = tmps.r1end;
						tmps.r1length = tmps.r1end - tmps.r1start;
						tmps.r2length = tmps.r2end - tmps.r2start;
						tmps.sl = tmps.r1length + tmps.r2length;
						i = i + 2;
						par.ReadBoundariers = tmps;
						std::cout << "Boundaries set to [" << par.ReadBoundariers.r1start << ":" << par.ReadBoundariers.r1end << "]" << std::endl;
						fSetBound = true;

						par.splitReads = false;
						std::cout << "Only two parameter given. Sequence will be treated as a single read." << std::endl;
						done++;
					}
				}
				if (!done){
					if (!finput) par.ReadBoundariers = Default_ReadBoundariers;
					std::cout << " WARNING: missing arguments. Boundaries set to default." << std::endl;
					Warning++;
				};
			};

			if (!std::string(argv[i]).compare("-er")) {
				std::cout << "Checking error rate file..." << std::flush;
				if (i + 1 <= argc) {
					par.erFileName = std::string(argv[i + 1]);
					std::ifstream erFile;
					erFile.open(par.erFileName, std::ios::in);
					if (!erFile)
					{
						std::cout << "Warning: Failed to open \"" << par.erFileName << "\". Error rate ignored." << std::endl;
						Warning++;
						er.given = false;
					}
					else {
						er.given = true;
						LoadErrorRate(erFile);
						std::cout << "done." << std::endl;

					};
					erFile.close();
					i++;
				}
				else {
					std::cout << "Warning: Missing argument. Error rate ignored." << std::endl;
					Warning++;
					er.given = false;
				};

				if (i + 1 <= argc) {
					par.er2FileName = std::string(argv[i + 1]);
					std::ifstream er2File;
					er2File.open(par.er2FileName, std::ios::in);
					if (er2File) {
						std::cout << "Checking 2nd error rate file..." << std::flush;
						LoadErrorRate2(er2File);
						std::cout << "done." << std::endl;


						i++;
					};
					er2File.close();
				};
			};

			if (!std::string(argv[i]).compare("-Qt")) {
				if (i + 1 <= argc) {
					if (argv[i + 1][0] != ('-')) {
						i++;
						if (has_only_digits(std::string(argv[i]))) {
							par.QThresh = atoi(argv[i]);
							std::cout << "Phred quality score threshold set to " << std::string(argv[i]) << "." << std::endl;
						}
						else {
							std::cout << "WARNING: \"" << std::string(argv[i]) << "\" is not a valid -Qt argument. Phred quality score threshold set to default." << std::endl;
							Warning++;
						};
					}
					else {			// no argument : next is -xx
						std::cout << "WARNING: Missing argument in -Qt. Phred quality score threshold set to default." << std::endl;
						Warning++;
					};
				}
				else {					// no argument : is the last argument
					std::cout << "WARNING: Missing argument in -Qt. Phred quality score threshold set to default." << std::endl;
					Warning++;
				};
			};
			if (!std::string(argv[i]).compare("-ert")) {
				if (i + 1 <= argc) {
					if (argv[i + 1][0] != ('-')) {
						i++;
						if (has_only_digits_float(std::string(argv[i]))) {
							par.erThresh = atof(argv[i]);
							std::cout << "Lambda-phage derived error rate threshold set to " << std::string(argv[i]) << "." << std::endl;
						}
						else {
							std::cout << "WARNING: \"" << std::string(argv[i]) << "\" is not a valid -ert argument. Lambda-phage derived error rate threshold set to default." << std::endl;
							Warning++;
						};
					}
					else {			// no argument : next is -xx
						std::cout << "WARNING: Missing argument in -ert. Lambda-phage derived error rate threshold set to default." << std::endl;
						Warning++;
					};
				}
				else {					// no argument : is the last argument
					std::cout << "WARNING: Missing argument in -ert. Lambda-phage derived error rate threshold set to default." << std::endl;
					Warning++;
				};
			};
			if (!std::string(argv[i]).compare("-test")) {
				if (i + 1 <= argc) {
					if (argv[i + 1][0] != ('-')) {
						i++;
						if (has_only_digits(std::string(argv[i]))) {
							par.MaxSequences = atoi(argv[i]);
							par.test = true;
							std::cout << "Test setting enabled. Only the first " << std::string(argv[i]) << " sequences will be read." << std::endl;
						}
						else {
							std::cout << "WARNING: \"" << std::string(argv[i]) << "\" is not a valid -test argument. The sequences which will be read are defaulted to the first 100 000." << std::endl;
							par.MaxSequences = 100000;
							Warning++;
						};
					}
					else {			// no argument : next is -xx
						std::cout << "Test setting enabled. The sequences which will be read are defaulted to the first 100 000." << std::endl;
						par.MaxSequences = 100000;
					};
				}
				else {					// no argument : is the last argument
					std::cout << "Test setting enabled. The sequences which will be read are defaulted to the first 100 000." << std::endl;
					par.MaxSequences = 100000;
				};
			};
			if (!std::string(argv[i]).compare("-j")) {
				par.splitReads = false;
				std::cout << "Fast setting enabled. Sequence will be treated as a single read." << std::endl;
			};
			if (!std::string(argv[i]).compare("-vl")) {
				par.vl = true;
				std::cout << "Variable length enabled. Sequence are allowed to be of different length." << std::endl;
			};
			if (!std::string(argv[i]).compare("-p")) {
				par.protan = true;
				std::cout << "Protein diversity analysis enabled." << std::endl;
			};
			if (!std::string(argv[i]).compare("-pfr")) {
				seedset++;
				if (i + 2 <= argc) {
					if (has_only_digits(std::string(argv[i + 1])) && has_only_digits(std::string(argv[i + 2]))) {
						par.r1frame = atoi(argv[i + 1]);
						par.r2frame = atoi(argv[i + 2]);
						i = i + 2;
						std::cout << "Protein frames set to [" << par.r1frame << ", " << par.r2frame << "." << std::endl;
						fSetBound = true;
					}
					else {
						if (!finput) par.ReadBoundariers = Default_ReadBoundariers;
						std::cout << " WARNING: \"" << (std::string(argv[i + 1])) << " " << (std::string(argv[i + 2])) << "\" are not valid arguments. Protein frames set to default." << std::endl;
						Warning++;
					};
				}
				else {
					if (!finput) par.ReadBoundariers = Default_ReadBoundariers;
					std::cout << " WARNING: missing arguments. Protein frames set to default." << std::endl;
					Warning++;
				};
			};

		};
		if (Warning > 0) {
			std::cout << "Warning Found: Resuming operation." << std::endl;
		};
		if (seedset > 1) {
			std::cout << "Seed set multiple times: Resuming operation." << std::endl;
		};
		if (par.vl) {
			if (er2.npos == 0) {
				std::cout << "Variable length input needs 2 error rate file, one for each read. Only 1 was provided. Ignoring error rate." << std::endl;
				er.given = false;
			};
		}
		else {
			if (er.npos != par.ReadBoundariers.sl) {
				std::cout << "Input file and error rate file have different sequence length: Ignoring error rate." << std::endl;
				er.given = false;
			};
		};


		if (fError == true) {
			std::cout << "Error Found: Aborting operation." << std::endl;
			exit(1);
		};
		if (finput > 1) {
			std::cout << "Input file given multiple times: Aborting operation." << std::endl;
			exit(1);
		};
		if (finput == 0) {
			std::cout << "Input file not given: Aborting operation." << std::endl;
			exit(1);
		};

		wholeSeq = (char *)malloc((par.ReadBoundariers.sl + 1) * sizeof(char));
		seedSeq = (char *)malloc((par.Seed.sl + 1) * sizeof(char));
		par.r1frame = (par.r1frame - 1) % 3;
		par.r2frame = (par.r2frame - 1) % 3;
		int protr1l = (par.ReadBoundariers.r1length - par.r1frame) / 3;
		int protr2l = (par.ReadBoundariers.r2length - par.r2frame) / 3;
		int protRl = (par.ReadBoundariers.sl - par.r1frame) / 3;

		if (par.vl) {
			par.R1frame = par.r1frame;
			par.R2frame = 0;
			DefaultParam Default_protBoundariers = { 0, protRl, 0, 0, protRl, protRl, 0 };
			par.ProtBound = Default_protBoundariers;
		}
		else {
			par.R1frame = par.r1frame;
			par.R2frame = ((par.ReadBoundariers.r1length - par.r1frame) % 3) + par.r1frame + par.r2frame;
			DefaultParam Default_protBoundariers = { 0, protr1l, protr1l, protr1l + protr2l, protr1l + protr2l, protr1l, protr2l };
			par.ProtBound = Default_protBoundariers;
		};

		if (par.Seed.r1length != 0) {
			if (par.Seed.r1start > par.ReadBoundariers.sl) {
				std::cout << "Seed (Read1) exceed read boundaries. Seed (Read1) will be ignored./n";
				par.Seed.r1start = 0;
				par.Seed.r1end = 0;
				par.Seed.r1length = 0;
			}
			else {
				if (par.Seed.r1end > par.ReadBoundariers.sl) {
					std::cout << "Seed (Read1) exceed read boundaries. Seed (Read1) will be ignored./n";
					par.Seed.r1start = 0;
					par.Seed.r1end = 0;
					par.Seed.r1length = 0;
				};
			};
		};

		if (par.Seed.r2length != 0) {
			if (par.Seed.r2start > par.ReadBoundariers.sl) {
				std::cout << "Seed (Read2) exceed read boundaries. Seed (Read2) will be ignored./n";
				par.Seed.r2start = par.Seed.r1end;
				par.Seed.r2end = par.Seed.r1end;
				par.Seed.r2length = 0;
			}
			else {
				if (par.Seed.r2end > par.ReadBoundariers.sl) {
					std::cout << "Seed (Read2) exceed read boundaries. Seed (Read2) will be ignored./n";
					par.Seed.r2start = par.Seed.r1end;
					par.Seed.r2end = par.Seed.r1end;
					par.Seed.r2length = 0;
				};
			};
		};
		if (par.Seed.r2length == 0 & par.Seed.r1length == 0) {
			std::cout << "No Seed given. Aborting operation" << std::endl;
			exit(1);
		}
		protSeq = (char *)malloc((par.ProtBound.sl + 1) * sizeof(char));
	}





	bool has_only_digits(std::string s)
	{
		return (s.find_first_not_of("0123456789") == std::string::npos);
	}
	bool has_only_digits_float(std::string s)
	{
		return (s.find_first_not_of("0123456789.") == std::string::npos);
	}

}