/*
Copyright 2014 Marco Fantini
This file is part of DEAL.
DEAL is free software : you can redistribute it and / or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DEAL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DEAL.If not, see <http://www.gnu.org/licenses/>.
*/

#include "SeqTree.h"

using namespace std;

namespace SeqTree
{

	void removenode(node *N) //rimuove nodi. ricorsivamente.
	{
		uint16_t nf = numfork(N);
		for (uint16_t i = 1; i <= nf; i++) removenode(N->Fork[i]); //perch il primo elemento  1
		free(N->Seq);
		free(N->Fpos);
		free(N->Fork);
		free(N);
	};

	void clonenode(node *destination, node *origin, uint16_t sl) //clona nodi. utile per la versione del programma che considera le basi incerte.
	{
		uint16_t NumFork = numfork(origin);
		destination->Fpos = (uint16_t *)malloc((NumFork + 2) * sizeof(uint16_t));
		destination->Fork = (node **)malloc((NumFork + 2) * sizeof(node *));
		destination->Seq = (char *)malloc((sl + 1) * sizeof(char));
		memcpy(destination->Fpos, origin->Fpos, (NumFork + 2)* sizeof(uint16_t));
		memcpy(destination->Fork, origin->Fork, (NumFork + 2)* sizeof(node *));
		memcpy(destination->Seq, origin->Seq, (sl + 1)* sizeof(char));
		//omp_init_lock(&(destination->lock));
	};

	void addfork(node *N, node *TREE, uint16_t sl, uint16_t i, uint16_t dimension) //crea una biforcazione
	{

		TREE->Fpos = (uint16_t *)realloc(TREE->Fpos, (dimension + 1) * sizeof(uint16_t));	//aggiungiamo uno spazio agli array Fpos 
		TREE->Fork = (node **)realloc(TREE->Fork, (dimension + 1) * sizeof(node *));		//e Fork per farci stare la nuova biforcazione
		uint16_t  newindex = indexfork_newindex(TREE, i);				//recuperiamo la posizione degli array in cui va inserito per mantenerlo ordinato ( posizione di biforcazione )

		memmove(TREE->Fpos + newindex + 1, TREE->Fpos + newindex, (dimension - newindex) * sizeof(uint16_t));
		memmove(TREE->Fork + newindex + 1, TREE->Fork + newindex, (dimension - newindex) * sizeof(node *));		
		TREE->Fpos[newindex] = i;
		TREE->Fork[newindex] = N;			//spostiamo i dati che lo seguono per non sovrascrivere niente e inseriamo i nuovi dati di posizione e * al nuovo ramo
		//omp_unset_lock(&(TREE->lock));

		char *newSeq = (char *)malloc((sl - i + 1) * sizeof(char));
		memcpy(newSeq, N->Seq + i, (sl - i + 1)* sizeof(char));
		free(N->Seq);
		N->Seq = newSeq;					//Ridimensioniamo la sequenza alla sola parte nuova.
		//omp_unset_lock(&(N->lock));
	};

	int insert(node *N, node *TREE, uint16_t sl, uint16_t i, uint16_t lastFp, int F)   //
	{
		//omp_set_lock(&(N->lock));
		//omp_set_lock(&(TREE->lock));
		while (i != sl)
		{
				if (N->Seq[i] == TREE->Seq[i - lastFp])	// se la base  compatibile alla Sequenza che sta scorrendo sull'albero
				{
					++i;								//passa alla base successiva
				}
				else{									//altrimenti guarda in quella posizione
					uint16_t indexf = indexfork(TREE, i);

					//se non trova niente restituisce l'ultimo indice dell'array (dimarray -1 perch parto da 0)
					//obv se non trova niente l'ultimo elemento ha un Fork[indexfork == dimarray -1] == NULL

					if (TREE->Fork[indexf])				//se esite gi una biforcazione in quel punto
					{							// (NULL = FALSE (vai a Else) TuttotranneNULL = TRUE (continua)
						//omp_unset_lock(&(TREE->lock));
						TREE = TREE->Fork[indexf];
						//omp_set_lock(&(TREE->lock));
						lastFp = i;
					}
					else{								//altrimenti creane una
						addfork(N, TREE, sl, i, indexf + 1);
						return 1;
					};
				};
		};
		//omp_unset_lock(&(N->lock));
		//omp_unset_lock(&(TREE->lock));

//#pragma omp critical										//arrivati all'ultima base della sequenza senza aver dovuto creare una nuova biforcazione
		weave((Seqnode *)(N->Fork[0]), (Seqnode *)TREE->Fork[0],F);			//intreccia la nuova sequenza al gruppo di sequenze preeesistenti che hai trovato con l'albero
		return 0;
	};

}