/*
Copyright 2014 Marco Fantini
This file is part of DEAL.
DEAL is free software : you can redistribute it and / or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DEAL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DEAL.If not, see <http://www.gnu.org/licenses/>.
*/


#include "SeqTree.h"
using namespace std;

namespace SeqTree
{

	void straighten(ofstream &oFile_group, ofstream &oFile_r1_group, ofstream &oFile_r2_group, ofstream &oFile_stat, ofstream &oFile_DF)
	{


		oFile_group << "#Groups\n";
		if (par.splitReads) oFile_r2_group << "#Reverse read2\n";
		if (par.splitReads) oFile_r1_group << "#Groups read1\n";

		Seqnode *L = SeqROOT;
		do
		{
			Seqnode *Seqi = L;
			++TOTSeq;
			if (!(TOTSeq % 10000)) cerr << "Calculating Groups and statistics..." << TOTSeq << "\r";
			if (Seqi->rp->LEFT_R != NULL) //se l'ho gi analizzata in precedenza.
			{
				oFile_group << "Group\t" << Seqi->ID << ":\n";
				while (Seqi->rp->LEFT_R != NULL) //fa il circolo e torna all'inizio
				{
					oFile_group << Seqi->ID << " ";
					Seqi->rp->LEFT_R = NULL;
					Seqi = Seqi->rp->RIGHT_R;
				};
					oFile_group << "\n";
			}

			if (par.splitReads){

				if (Seqi->rp->LEFT_r1 != NULL) //se l'ho gi analizzata in precedenza.
				{
					oFile_r1_group << "Group\t" << Seqi->ID << ":\n";
					while (Seqi->rp->LEFT_r1 != NULL)
					{
						oFile_r1_group << Seqi->ID << " ";
						Seqi->rp->LEFT_r1 = NULL;
						Seqi = Seqi->rp->RIGHT_r1;
					};
					oFile_r1_group << "\n";
				};

				if (Seqi->rp->LEFT_r2 != NULL) //se l'ho gi analizzata in precedenza.
				{
					oFile_r2_group << "Group\t" << Seqi->ID << ":\n";
					while (Seqi->rp->LEFT_r2 != NULL)
					{
						oFile_r2_group << Seqi->ID << " ";
						Seqi->rp->LEFT_r2 = NULL;
						Seqi = Seqi->rp->RIGHT_r2;
					};
					oFile_r2_group << "\n";
				};


			};

			free(Seqi->rp);
			if (!par.splitReads) oFile_DF << TOTSeq << "\t" << Seqi->Group_R->ID << "\n";
			if ( par.splitReads) oFile_DF << TOTSeq << "\t" << Seqi->Group_R->ID << "\t" << Seqi->Group_r1->ID << "\t" << Seqi->Group_r2->ID << "\n";
			L = L->next;
		} while (L != NULL);


		cout << "Calculating Groups and statistics...Done!        \n";
		cout << "Seed statistic:\n";
		unsigned long cumSeq = 0;

		if (par.splitReads){
			cumSeq = 0;
			cout << "Total sequence number = " << TOTSeq << "\nTotal groups number read1 = " << TOTGroup_r1 << "\n";
			oFile_stat << "#Group statistic read1\n";
			oFile_stat << "Total sequence number = " << TOTSeq << "\nTotal groups number read1 = " << TOTGroup_r1 << "\n";
			for (std::map<unsigned long, unsigned long>::iterator it = ngroup_r1.begin(); it != ngroup_r1.end(); ++it){
				if (it->first < 20) cout << "Group r1[" << it->first << "] = " << it->second << "\n";
				oFile_stat << "Group\tr1\t" << it->first << "\t" << it->second << '\n';
				cumSeq += it->first * it->second;
			};

			cumSeq = 0;
			cout << "Total sequence number = " << TOTSeq << "\nTotal groups number read2 = " << TOTGroup_r2 << "\n";
			oFile_stat << "#Group statistic read2\n";
			oFile_stat << "Total sequence number = " << TOTSeq << "\nTotal groups number read2 = " << TOTGroup_r2 << "\n";
			for (std::map<unsigned long, unsigned long>::iterator it = ngroup_r2.begin(); it != ngroup_r2.end(); ++it){
				if (it->first < 20) cout << "Group r2[" << it->first << "] = " << it->second << "\n";
				oFile_stat << "Group\tr2\t" << it->first << "\t" << it->second << '\n';
				cumSeq += it->first * it->second;
			};
		};

		cumSeq = 0;
		cout << "Total sequence number = " << TOTSeq << "\nTotal groups number = " << TOTGroup_R << "\n";
		oFile_stat << "#Group statistic\n";
		oFile_stat << "Total sequence number = " << TOTSeq << "\nTotal groups number = " << TOTGroup_R << "\n";
		for (std::map<unsigned long, unsigned long>::iterator it = ngroup_R.begin(); it != ngroup_R.end(); ++it){
			if (it->first < 20) cout << "Group R[" << it->first << "] = " << it->second << "\n";
			oFile_stat << "Group\tR\t" << it->first << "\t" << it->second << '\n';
			cumSeq += it->first * it->second;
		};

	};

	void weave(Seqnode *S1, Seqnode *S2, int F) //nota S1 sono le nuove sequenze, S2 quelle derivate dall'albero
	{
		 if (F == 0){
			Seqnode *S1L = S1->rp->LEFT_R;
			Seqnode *S2L = S2->rp->LEFT_R;

			S1L->rp->RIGHT_R = S2;
			S2->rp->LEFT_R = S1L;

			S2L->rp->RIGHT_R = S1;
			S1->rp->LEFT_R = S2L;

			ngroup_R[S2->Group_R->TotCount]--;
			S2->Group_R->TotCount += S1->Group_R->TotCount;
			ngroup_R[S1->Group_R->TotCount]--;
			ngroup_R[S2->Group_R->TotCount]++;
			free(S1->Group_R->DefSeq);
			free(S1->Group_R);
			S1->Group_R = S2->Group_R;
			TOTGroup_R--;

		} else if (F==1){
			Seqnode *S1L = S1->rp->LEFT_r1;
			Seqnode *S2L = S2->rp->LEFT_r1;

			S1L->rp->RIGHT_r1 = S2;
			S2->rp->LEFT_r1 = S1L;

			S2L->rp->RIGHT_r1 = S1;
			S1->rp->LEFT_r1 = S2L;

			ngroup_r1[S2->Group_r1->TotCount]--;
			S2->Group_r1->TotCount += S1->Group_r1->TotCount;
			ngroup_r1[S1->Group_r1->TotCount]--;
			ngroup_r1[S2->Group_r1->TotCount]++;
			free(S1->Group_r1->DefSeq);
			free(S1->Group_r1);
			S1->Group_r1 = S2->Group_r1;
			TOTGroup_r1--;

		} else if (F == 2){
			Seqnode *S1L = S1->rp->LEFT_r2;
			Seqnode *S2L = S2->rp->LEFT_r2;

			S1L->rp->RIGHT_r2 = S2;
			S2->rp->LEFT_r2 = S1L;

			S2L->rp->RIGHT_r2 = S1;
			S1->rp->LEFT_r2 = S2L;

			ngroup_r2[S2->Group_r2->TotCount]--;
			S2->Group_r2->TotCount += S1->Group_r2->TotCount;
			ngroup_r2[S1->Group_r2->TotCount]--;
			ngroup_r2[S2->Group_r2->TotCount]++;
			free(S1->Group_r2->DefSeq);
			free(S1->Group_r2);
			S1->Group_r2 = S2->Group_r2;
			TOTGroup_r2--;

		};
	};
}