/*
Copyright 2014 Marco Fantini
This file is part of DEAL.
DEAL is free software : you can redistribute it and / or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DEAL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with DEAL.If not, see <http://www.gnu.org/licenses/>.
*/


#include "SeqTree.h"

using namespace std;

namespace SeqTree
{

	char recprot(char b[3], char hist, int lvl)
	{
		if (lvl < 3){
			char tmp = b[lvl];
			for (int i = 0; i < 4; i++){
				if (tmp & (1 << i)) {
					b[lvl] = (tmp & (1 << i));
					char newtrans = recprot(b, hist, lvl + 1);
					b[lvl] = tmp;
					if (newtrans == 'O') return 'O';
					if (hist != 0) {
						if (newtrans != hist) {
							if (newtrans == '*' || hist == '*') {
								hist = 'O';
								return 'O';
							}
							else {
								hist = 'X';
							}
						}
					}
					else {
						hist = newtrans;
					}


				}
			}
			return hist;
		}
		else {

			return prot[b[0]][b[1]][b[2]];
		}
	}

	void loadprot(){
		const unsigned char A = 1;
		const unsigned char C = 2;
		const unsigned char G = 4;
		const unsigned char T = 8;

		int basebp = 4;
		int extbp = 16;

		prot = new char**[extbp*2]();
		for (int i = 0; i < extbp*2; i++)
		{
			prot[i] = new char*[extbp*2]();
			for (int j = 0; j < extbp*2; j++)
			{
				prot[i][j] = new char[extbp*2]();
			}
		}

		prot[T][T][T] = 'F';
		prot[T][T][C] = 'F';
		prot[T][T][A] = 'L';
		prot[T][T][G] = 'L';
		prot[C][T][T] = 'L';
		prot[C][T][C] = 'L';
		prot[C][T][A] = 'L';
		prot[C][T][G] = 'L';
		prot[A][T][T] = 'I';
		prot[A][T][C] = 'I';
		prot[A][T][A] = 'I';
		prot[A][T][G] = 'M';
		prot[G][T][T] = 'V';
		prot[G][T][C] = 'V';
		prot[G][T][A] = 'V';
		prot[G][T][G] = 'V';
		prot[T][C][T] = 'S';
		prot[T][C][C] = 'S';
		prot[T][C][A] = 'S';
		prot[T][C][G] = 'S';
		prot[C][C][T] = 'P';
		prot[C][C][C] = 'P';
		prot[C][C][A] = 'P';
		prot[C][C][G] = 'P';
		prot[A][C][T] = 'T';
		prot[A][C][C] = 'T';
		prot[A][C][A] = 'T';
		prot[A][C][G] = 'T';
		prot[G][C][T] = 'A';
		prot[G][C][C] = 'A';
		prot[G][C][A] = 'A';
		prot[G][C][G] = 'A';
		prot[T][A][T] = 'Y';
		prot[T][A][C] = 'Y';
		prot[T][A][A] = '*';
		prot[T][A][G] = '*';
		prot[C][A][T] = 'H';
		prot[C][A][C] = 'H';
		prot[C][A][A] = 'Q';
		prot[C][A][G] = 'Q';
		prot[A][A][T] = 'N';
		prot[A][A][C] = 'N';
		prot[A][A][A] = 'K';
		prot[A][A][G] = 'K';
		prot[G][A][T] = 'D';
		prot[G][A][C] = 'D';
		prot[G][A][A] = 'E';
		prot[G][A][G] = 'E';
		prot[T][G][T] = 'C';
		prot[T][G][C] = 'C';
		prot[T][G][A] = '*';
		prot[T][G][G] = 'W';
		prot[C][G][T] = 'R';
		prot[C][G][C] = 'R';
		prot[C][G][A] = 'R';
		prot[C][G][G] = 'R';
		prot[A][G][T] = 'S';
		prot[A][G][C] = 'S';
		prot[A][G][A] = 'R';
		prot[A][G][G] = 'R';
		prot[G][G][T] = 'G';
		prot[G][G][C] = 'G';
		prot[G][G][A] = 'G';
		prot[G][G][G] = 'G';

		for (int i = 1; i < extbp; i++)
		{
			for (int j = 1; j < extbp; j++)
			{
				for (int k = 1; k < extbp; k++)
				{
					char b[3] = { b[0] = i, b[1] = j, b[2] = k };
					prot[i][j][k] = recprot(b, 0, 0);
					prot[i + inc][j][k] = prot[i][j][k];
					prot[i][j + inc][k] = prot[i][j][k];
					prot[i][j][k + inc] = prot[i][j][k];
					prot[i + inc][j + inc][k] = prot[i][j][k];
					prot[i + inc][j][k + inc] = prot[i][j][k];
					prot[i][j + inc][k + inc] = prot[i][j][k];
					prot[i + inc][j + inc][k + inc] = prot[i][j][k];
					//cout << Decompile[i] << Decompile[j] << Decompile[k] << "  " << prot[i][j][k] << std::endl;  //debug!
				}
			}
		}

	}

	void ProteinParseSequence(Groupnode *GN, int F)
	{
		char *Sequence = GN->DefSeq;
		SeqnodeLRpointers *new_rp;

		node *newnode_R = NULL;
		node *newnode_r1 = NULL;
		node *newnode_r2 = NULL;
		static Seqnode *oldSnode_R = NULL;
		static Seqnode *oldSnode_r1 = NULL;
		static Seqnode *oldSnode_r2 = NULL;
		Seqnode *newSnode = NULL;
		static Groupnode *oldGnode_R;
		static Groupnode *newGnode_R;
		static Groupnode *oldGnode_r1;
		static Groupnode *newGnode_r1;
		static Groupnode *oldGnode_r2;
		static Groupnode *newGnode_r2;

		bool pSTOPflag = 0;
		if (F == 0){
			translateseq(Sequence, protSeq, &pSTOPflag, F);

			if (!pSTOPflag) {
				TOTSeq_R_nopSTOP += GN->TotCount;
				TOTGroup_R_nopSTOP++;
			}

			newGnode_R = (Groupnode *)malloc(sizeof(Groupnode));
			newGnode_R->ID = GN->ID;
			newGnode_R->Depth = GN->Depth;
			newGnode_R->count = 0; 
			newGnode_R->DefSeq = (char *)calloc(par.ProtBound.sl + 1, sizeof(char));
			memcpy(newGnode_R->DefSeq, protSeq, par.ProtBound.sl * sizeof(char));
			newGnode_R->DefSeq[par.ProtBound.sl] = 0;
			newGnode_R->next = NULL;
			newGnode_R->SubGroup = NULL;
			newGnode_R->TotCount = GN->TotCount;
			ngroup_R[newGnode_R->TotCount]++;
			TOTGroup_R++;

			newSnode = (Seqnode *)malloc(sizeof(Seqnode));
			newSnode->next = NULL;
			newSnode->Group_R = newGnode_R;
			newSnode->ID = GN->ID;
			new_rp = (SeqnodeLRpointers *)malloc(sizeof(SeqnodeLRpointers));
			newSnode->rp = new_rp;
			newSnode->rp->RIGHT_R = newSnode;
			newSnode->rp->LEFT_R = newSnode;
			newSnode->Group_R = newGnode_R;
			GN->protein = newSnode;

			newnode_R = (node *)malloc(sizeof(node));
			newnode_R->Fpos = (uint16_t *)malloc(2 * sizeof(uint16_t)); //1+1
			newnode_R->Fork = (node **)malloc(2 * sizeof(node *));
			newnode_R->Seq = (char *)calloc(par.ProtBound.sl + 1, sizeof(char));
			newnode_R->Fpos[0] = 9999;
			newnode_R->Fpos[1] = 0;
			newnode_R->Fork[0] = (node *)newSnode;
			newnode_R->Fork[1] = NULL;

			memcpy(newnode_R->Seq, protSeq, par.ProtBound.sl * sizeof(char));
			newnode_R->Seq[par.ProtBound.sl] = 0;

			if (ROOT_R){ // != NULL
				insert(newnode_R, ROOT_R, par.ProtBound.sl, 0, 0, F);
			}else {
				pSeqROOT_R = newSnode;
				ROOT_R = newnode_R;
				pGROOT_R = newGnode_R;
				oldSnode_R = pSeqROOT_R;
				oldGnode_R = pGROOT_R;
			};
													// newSnode pu essere stato modificato da insert ( il primo elemento di newnode)
													// newSnode->Group_R diverso da newGnode_R vuole dire che ha trovato una proteina con la stessa sequenza, e quindi non va inserito come next
			if (newGnode_R == newSnode->Group_R){ // = se non  cambiato il novo gruppo... (ovvero se  cambiato vuole dire che ha trovato una proteina con la stessa sequenza e quindi non hanno pi lo stesso puntatore.)
				if (newGnode_R != pGROOT_R) oldGnode_R->next = newGnode_R;	//ROOT non nexta se stessa!
				oldGnode_R = newGnode_R;
			};

			if (newSnode != pSeqROOT_R) oldSnode_R->next = newSnode; //per lo stesso principio del seed la ROOT non deve nextare se stessa!  come le normali sequenze (che hanno next vuoto)
			oldSnode_R = newSnode;
		};
		if (F == 1){

			translateseq(Sequence, protSeq, &pSTOPflag, F);

			if (!pSTOPflag) {
				TOTSeq_r1_nopSTOP += GN->TotCount;
				TOTGroup_r1_nopSTOP++;
			}

			newGnode_r1 = (Groupnode *)malloc(sizeof(Groupnode));
			newGnode_r1->ID = GN->ID;
			newGnode_r1->Depth = GN->Depth;
			newGnode_r1->count = 0;
			newGnode_r1->DefSeq = (char *)calloc(par.ProtBound.r1length + 1, sizeof(char));
			memcpy(newGnode_r1->DefSeq, protSeq, par.ProtBound.r1length * sizeof(char));
			newGnode_r1->DefSeq[par.ProtBound.r1length] = 0;
			newGnode_r1->next = NULL;
			newGnode_r1->SubGroup = NULL;
			newGnode_r1->TotCount = GN->TotCount;
			ngroup_r1[newGnode_r1->TotCount]++;
			TOTGroup_r1++;

			newSnode = (Seqnode *)malloc(sizeof(Seqnode));
			newSnode->next = NULL;
			newSnode->Group_r1 = newGnode_r1;
			newSnode->ID = GN->ID;
			new_rp = (SeqnodeLRpointers *)malloc(sizeof(SeqnodeLRpointers));
			newSnode->rp = new_rp;
			newSnode->rp->RIGHT_r1 = newSnode;
			newSnode->rp->LEFT_r1 = newSnode;
			newSnode->Group_r1 = newGnode_r1;
			GN->protein = newSnode;

			newnode_r1 = (node *)malloc(sizeof(node));
			newnode_r1->Fpos = (uint16_t *)malloc(2 * sizeof(uint16_t)); //1+1
			newnode_r1->Fork = (node **)malloc(2 * sizeof(node *));
			newnode_r1->Seq = (char *)calloc(par.ProtBound.r1length + 1, sizeof(char));
			newnode_r1->Fpos[0] = 9999;
			newnode_r1->Fpos[1] = 0;
			newnode_r1->Fork[0] = (node *)newSnode;
			newnode_r1->Fork[1] = NULL;

			memcpy(newnode_r1->Seq, protSeq, par.ProtBound.r1length * sizeof(char));
			newnode_r1->Seq[par.ProtBound.r1length] = 0;

			if (ROOT_r1){ // != NULL
				insert(newnode_r1, ROOT_r1, par.ProtBound.r1length, 0, 0, F);
			}
			else {
				pSeqROOT_r1 = newSnode;
				ROOT_r1 = newnode_r1;
				pGROOT_r1 = newGnode_r1;
				oldSnode_r1 = pSeqROOT_r1;
				oldGnode_r1 = pGROOT_r1;
			};

			if (newGnode_r1 == newSnode->Group_r1){ // = se non  cambiato il novo gruppo
				if (newGnode_r1 != pGROOT_r1) oldGnode_r1->next = newGnode_r1;	//ROOT non nexta se stessa!
				oldGnode_r1 = newGnode_r1;
			};

			if (newSnode != pSeqROOT_r1) oldSnode_r1->next = newSnode; //per lo stesso principio del seed la ROOT non deve nextare se stessa!  come le normali sequenze (che hanno next vuoto)
			oldSnode_r1 = newSnode;
		};
		if (F == 2){

			translateseq(Sequence, protSeq, &pSTOPflag, F);

			if (!pSTOPflag) {
				TOTSeq_r2_nopSTOP += GN->TotCount;
				TOTGroup_r2_nopSTOP++;
			}

			newGnode_r2 = (Groupnode *)malloc(sizeof(Groupnode));
			newGnode_r2->ID = GN->ID;
			newGnode_r2->Depth = GN->Depth;
			newGnode_r2->count = 0;
			newGnode_r2->DefSeq = (char *)calloc(par.ProtBound.r2length + 1, sizeof(char));
			memcpy(newGnode_r2->DefSeq, protSeq, par.ProtBound.r2length * sizeof(char));
			newGnode_r2->DefSeq[par.ProtBound.r2length] = 0;
			newGnode_r2->next = NULL;
			newGnode_r2->SubGroup = NULL;
			newGnode_r2->TotCount = GN->TotCount;
			ngroup_r2[newGnode_r2->TotCount]++;
			TOTGroup_r2++;

			newSnode = (Seqnode *)malloc(sizeof(Seqnode));
			newSnode->next = NULL;
			newSnode->Group_r2 = newGnode_r2;
			newSnode->ID = GN->ID;
			new_rp = (SeqnodeLRpointers *)malloc(sizeof(SeqnodeLRpointers));
			newSnode->rp = new_rp;
			newSnode->rp->RIGHT_r2 = newSnode;
			newSnode->rp->LEFT_r2 = newSnode;
			newSnode->Group_r2 = newGnode_r2;
			GN->protein = newSnode;

			newnode_r2 = (node *)malloc(sizeof(node));
			newnode_r2->Fpos = (uint16_t *)malloc(2 * sizeof(uint16_t)); //1+1
			newnode_r2->Fork = (node **)malloc(2 * sizeof(node *));
			newnode_r2->Seq = (char *)calloc(par.ProtBound.r2length + 1, sizeof(char));
			newnode_r2->Fpos[0] = 9999;
			newnode_r2->Fpos[1] = 0;
			newnode_r2->Fork[0] = (node *)newSnode;
			newnode_r2->Fork[1] = NULL;

			memcpy(newnode_r2->Seq, protSeq, par.ProtBound.r2length * sizeof(char));
			newnode_r2->Seq[par.ProtBound.r2length] = 0;

			if (ROOT_r2){ // != NULL
				insert(newnode_r2, ROOT_r2, par.ProtBound.r2length, 0, 0, F);
			}
			else {
				pSeqROOT_r2 = newSnode;
				ROOT_r2 = newnode_r2;
				pGROOT_r2 = newGnode_r2;
				oldSnode_r2 = pSeqROOT_r2;
				oldGnode_r2 = pGROOT_r2;
			};

			if (newGnode_r2 == newSnode->Group_r2){ // = se non  cambiato il novo gruppo
				if (newGnode_r2 != pGROOT_r2) oldGnode_r2->next = newGnode_r2;	//ROOT non nexta se stessa!
				oldGnode_r2 = newGnode_r2;
			};

			if (newSnode != pSeqROOT_r2) oldSnode_r2->next = newSnode; //per lo stesso principio del seed la ROOT non deve nextare se stessa!  come le normali sequenze (che hanno next vuoto)
			oldSnode_r2 = newSnode;
		};

	};


	void proteinanalysis()
	{

		string ifilename = par.iFileName;

		string ofilenameseq_r1;
		string ofilenameseq_r2;
		string ofilenameseq_R;
		ofstream ofileseq_r1;
		ofstream ofileseq_r2;
		ofstream ofileseq_R;

		if (par.splitReads){
			ofilenameseq_r1 = string(ifilename);
			ofilenameseq_r1 += "_Protein_Group_r1_seq.txt";
			cout << "output: " << ofilenameseq_r1.c_str() << '\n';
			ofileseq_r1.open(ofilenameseq_r1.c_str(), ios::out);

			ofilenameseq_r2 = string(ifilename);
			ofilenameseq_r2 += "_Protein_Group_r2_seq.txt";
			cout << "output: " << ofilenameseq_r2.c_str() << '\n';
			ofileseq_r2.open(ofilenameseq_r2.c_str(), ios::out);

		};

		ofilenameseq_R = string(ifilename);
		ofilenameseq_R += "_Protein_Group_seq.txt";
		cout << "output: " << ofilenameseq_R.c_str() << '\n';
		ofileseq_R.open(ofilenameseq_R.c_str(), ios::out);

		ofstream ofile_stat;
		string ofilename_stat;
		ofilename_stat = string(ifilename);
		ofilename_stat += "_Stat_Protein.txt";
		cout << "output: " << ofilename_stat.c_str() << '\n';
		ofile_stat.open(ofilename_stat.c_str(), ios::out);

		ofstream ofile_group_protein;
		string ofilename_group_protein;
		ofilename_group_protein = string(ifilename);
		ofilename_group_protein += "_Protein_Group_seq.txt";
		cout << "output: " << ofilename_group_protein.c_str() << '\n';
		ofile_group_protein.open(ofilename_group_protein.c_str(), ios::out);

		ofstream ofile_r1_group_protein;
		string ofilename_r1_group_protein;
		ofstream ofile_r2_group_protein;
		string ofilename_r2_group_protein;

		if (par.splitReads){
			ofilename_r1_group_protein = string(ifilename);
			ofilename_r1_group_protein += "_Protein_Group_r1_seq.txt";
			cout << "output: " << ofilename_r1_group_protein.c_str() << '\n';
			ofile_r1_group_protein.open(ofilename_r1_group_protein.c_str(), ios::out);

			ofilename_r2_group_protein = string(ifilename);
			ofilename_r2_group_protein += "_Protein_Group_r1_seq.txt";
			cout << "output: " << ofilename_r2_group_protein.c_str() << '\n';
			ofile_r2_group_protein.open(ofilename_r2_group_protein.c_str(), ios::out);
		};

		long TOTprotG = -1;
		Groupnode *G = pGROOT_R;
		do
		{
			Groupnode *Groupi = G;
			TOTprotG++;
			if (!(TOTprotG % 10000)) cout << "Calculating Groups and statistics..." << TOTprotG << "\r";
			ofile_group_protein << ">Group:" << TOTprotG << "x" << G->TotCount << ":\n"; //es >Group:0x1
			ofile_group_protein << G->DefSeq << "\n";
			G = G->next;
		} while (G != NULL);
		if (par.splitReads){
			TOTprotG = -1;
			G = pGROOT_r1;
			do
			{
				Groupnode *Groupi = G;
				TOTprotG++;
				if (!(TOTprotG % 10000)) cout << "Calculating Groups and statistics..." << TOTprotG << "\r";
				ofile_r1_group_protein << ">Group:" << TOTprotG << "x" << G->TotCount << ":\n"; //es >Group:0x1
				ofile_r1_group_protein << G->DefSeq << "\n";
				G = G->next;
			} while (G != NULL);
			TOTprotG = -1;
			G = pGROOT_r2;
			do
			{
				Groupnode *Groupi = G;
				TOTprotG++;
				if (!(TOTprotG % 10000)) cout << "Calculating Groups and statistics..." << TOTprotG << "\r";
				ofile_r2_group_protein << ">Group:" << TOTprotG << "x" << G->TotCount << ":\n"; //es >Group:0x1
				ofile_r2_group_protein << G->DefSeq << "\n";
				G = G->next;
			} while (G != NULL);
		};
		TOTprotG++;
		unsigned long cumSeq = 0;
		cout << "Calculating Groups and statistics...Done!                    \n";
		cout << "Out of " << TOTSeq << " proteins, " << TOTSeq_R_nopSTOP << " (" << (double)((unsigned long)(10000 * (TOTSeq_R_nopSTOP /TOTSeq))) / 100 << "%) had no STOP codon\n";
		cout << "Out of " << TOTGroup_R_2ndstep << " groups, " << TOTGroup_R_nopSTOP << " (" << (double)((unsigned long)(10000 * (TOTGroup_R_nopSTOP / TOTGroup_R_2ndstep))) / 100 << "%) had no STOP codon\n";
		if (par.splitReads) {
			cout << "Out of " << TOTSeq << " r1 proteins, " << TOTSeq_r1_nopSTOP << " (" << (double)((unsigned long)(10000 * (TOTSeq_r1_nopSTOP / TOTSeq))) / 100 << "%) had no STOP codon\n";
			cout << "Out of " << TOTGroup_r1_2ndstep << " r1 groups, " << TOTGroup_r1_nopSTOP << " (" << (double)((unsigned long)(10000 * (TOTGroup_r1_nopSTOP / TOTGroup_r1_2ndstep))) / 100 << "%) had no STOP codon\n";
			cout << "Out of " << TOTSeq << " r2 proteins, " << TOTSeq_r2_nopSTOP << " (" << (double)((unsigned long)(10000 * (TOTSeq_r2_nopSTOP / TOTSeq))) / 100 << "%) had no STOP codon\n";
			cout << "Out of " << TOTGroup_r2_2ndstep << " r2 groups, " << TOTGroup_r2_nopSTOP << " (" << (double)((unsigned long)(10000 * (TOTGroup_r2_nopSTOP / TOTGroup_r2_2ndstep))) / 100 << "%) had no STOP codon\n";
		}
		cout << "Protein Information: (" << TOTprotG << ")\n";

		if (par.splitReads){
			cumSeq = 0;
			cout << "Total sequences number = " << TOTSeq << "\nTotal groups number read1 = " << TOTGroup_r1 << "\n";
			ofile_stat << "#Group statistic read1\n";
			ofile_stat << "Total sequences number = " << TOTSeq << "\nTotal groups number read1 = " << TOTGroup_r1 << "\n";
			for (std::map<unsigned long, unsigned long>::iterator it = ngroup_r1.begin(); it != ngroup_r1.end(); ++it){
				if (it->first < 20) cout << "Group r1[" << it->first << "] = " << it->second << "\n";
				ofile_stat << "Group\tr1\t" << it->first << "\t" << it->second << '\n';
				cumSeq += it->first * it->second;
			};

			cumSeq = 0;
			cout << "Total sequences number = " << TOTSeq << "\nTotal groups number read2 = " << TOTGroup_r2 << "\n";
			ofile_stat << "#Group statistic read2\n";
			ofile_stat << "Total sequences number = " << TOTSeq << "\nTotal groups number read2 = " << TOTGroup_r2 << "\n";
			for (std::map<unsigned long, unsigned long>::iterator it = ngroup_r2.begin(); it != ngroup_r2.end(); ++it){
				if (it->first < 20) cout << "Group r2[" << it->first << "] = " << it->second << "\n";
				ofile_stat << "Group\tr2\t" << it->first << "\t" << it->second << '\n';
				cumSeq += it->first * it->second;
			};
		};

		cumSeq = 0;
		cout << "Total sequences number = " << TOTSeq << "\nTotal groups number = " << TOTGroup_R << "\n";
		ofile_stat << "#Group statistic\n";
		ofile_stat << "Total sequences number = " << TOTSeq << "\nTotal groups number = " << TOTGroup_R << "\n";
		for (std::map<unsigned long, unsigned long>::iterator it = ngroup_R.begin(); it != ngroup_R.end(); ++it){
			if (it->first < 20) cout << "Group R[" << it->first << "] = " << it->second << "\n";
			ofile_stat << "Group\tR\t" << it->first << "\t" << it->second << '\n';
			cumSeq += it->first * it->second;
		};

	};

	void translateseq(char* Sequence, char* destination, bool* pSTOPflag, int F)
	{
		if (F == 0){
			for (unsigned int i = par.ProtBound.r1start; i < par.ProtBound.r1end; ++i){
				destination[i] = prot[0x0f & Sequence[(i * 3) + par.R1frame]][0x0f & Sequence[(i * 3) + 1 + par.R1frame]][0x0f & Sequence[(i * 3) + 2 + par.R1frame]];
				if (destination[i] == '*' | destination[i] == 'O') 	*pSTOPflag = true;
			};

			for (unsigned int i = par.ProtBound.r2start; i <par.ProtBound.r2end; ++i){

				destination[i] = prot[Sequence[(i * 3) + par.R2frame]][Sequence[(i * 3) + 1 + par.R2frame]][Sequence[(i * 3) + 2 + par.R2frame]];
			};
			destination[par.ProtBound.sl] = 0;
		};
		if (F == 1){
			for (unsigned int i = par.ProtBound.r1start; i < par.ProtBound.r1end; ++i){
				destination[i] = prot[Sequence[(i * 3) + par.r1frame]][Sequence[(i * 3) + 1 + par.r1frame]][Sequence[(i * 3) + 2 + par.r1frame]];
				if (destination[i] == '*' | destination[i] == 'O') 	*pSTOPflag = true;
			};
			destination[par.ProtBound.r1length] = 0;
		};
		if (F == 2){
			int k = 0;
			for (unsigned int i = par.ProtBound.r2start; i <par.ProtBound.r2end; ++i){
				destination[k] = prot[Sequence[(k * 3) + par.r2frame]][Sequence[(k * 3) + 1 + par.r2frame]][Sequence[(k * 3) + 2 + par.r2frame]];
				k++;
				if (destination[k] == '*' | destination[k] == 'O') 	*pSTOPflag = true;
			};
			destination[par.ProtBound.r2length] = 0;
		};
	};

}
